"""
### Perform convolutions on arrays.  ###
"""

# External
import numpy as np
from numba import jit, prange


def weight_distance(arr, method, decay, sigma, center, spherical, radius):
    """ Weight the kernel by distance. """

    if center == 0.0:
        normed = np.linalg.norm(arr)
    else:
        normed = np.linalg.norm(arr - np.array([0, 0, center]))

    if normed == 0.0:
        weight = 1.0

    if method is None:
        weight = 1.0
    elif method == "linear":
        weight = np.power((1 - decay), normed)
    elif method == "sqrt":
        weight = np.power(np.sqrt((1 - decay)), normed)
    elif method == "power":
        weight = np.power(np.power((1 - decay), 2), normed)
    elif method == "log":
        weight = np.log(normed + 2)
    elif method == "gaussian":
        weight = np.exp(-(np.power(normed, 2)) / (2 * np.power(sigma, 2)))
    else:
        raise ValueError("Unable to parse parameters distance_calc.")

    if spherical:
        sqrt_2 = np.sqrt(2)
        half_sqrt_2 = np.divide(sqrt_2, 2)

        if normed > (radius - half_sqrt_2) and normed < (radius + half_sqrt_2):
            adjustment = sqrt_2 / normed
        elif normed > (radius - half_sqrt_2):
            adjustment = 0.0
        elif normed < (radius + half_sqrt_2):
            adjustment = 1.0
        else:
            adjustment = 1.0

        return weight * adjustment

    return weight


def rotate_kernel(bottom_right):
    """ Create a whole kernel from a quadrant. """

    size = ((bottom_right.shape[0] - 1) * 2) + 1
    depth = bottom_right.shape[2]
    kernel = np.zeros((size, size, depth), dtype="float32")

    top_right = np.flipud(bottom_right)
    lower_left = np.fliplr(bottom_right)
    top_left = np.flipud(lower_left)

    kernel[size // 2:, size // 2:, :] = bottom_right
    kernel[:1 + -size // 2, :1 + -size // 2, :] = top_left
    kernel[1 + size // 2:, :size // 2, :] = lower_left[1:, :-1, :]
    kernel[:size // 2, 1 + size // 2:, :] = top_right[:-1, 1:, :]

    return kernel


def get_kernel(
    size,
    depth=1,
    hole=False,
    inverted=False,
    normalise=True,
    multi_dimensional=False,
    multi_dimensional_center=0,
    spherical=False,
    distance_weight=None,
    distance_decay=0.2,
    distance_sigma=1,
):
    """ Get a square kernel for convolutions.
        returns: kernel, weights, offsets
    """

    assert size >= 3, "Kernel must have atleast size 3."
    assert size % 2 != 0, "Kernel must be an uneven size."
    assert isinstance(size, int), "Kernel must be an integer."
    assert depth >= 1, "Depth must be a positive integer"
    assert isinstance(depth, int), "Depth must be an integer."

    if distance_weight is False:
        distance_weight = None

    quadrant = np.zeros((1 + size // 2, 1 + size // 2, depth), dtype="float32")

    for idx_x in range(0, quadrant.shape[0]):
        for idx_y in range(0, quadrant.shape[1]):
            for idx_z in range(0, quadrant.shape[2]):

                z_value = idx_z if multi_dimensional else 0

                weighted = weight_distance(
                    np.array([idx_x, idx_y, z_value], dtype="float32"),
                    method=distance_weight,
                    decay=distance_decay,
                    sigma=distance_sigma,
                    center=multi_dimensional_center,
                    spherical=spherical,
                    radius=size // 2,
                )

                quadrant[idx_x, idx_y, idx_z] = weighted
    if hole:
        for idx_z in range(0, quadrant.shape[2]):
            quadrant[0, 0, idx_z] = 0

    kernel = rotate_kernel(quadrant)

    if distance_weight == "log":
        kernel = kernel.max() - kernel

    if inverted:
        kernel = 1 - kernel

    if normalise:
        if multi_dimensional:
            summed = kernel.sum()
            if summed != 0.0:
                kernel = kernel / summed
        else:
            summed = kernel.sum(axis=(0, 1))

            for dim in range(0, depth):
                kernel[:, :, dim] = kernel[:, :, dim] / summed[dim]

    weights = []
    offsets = []

    for idx_x in range(0, kernel.shape[0]):
        for idx_y in range(0, kernel.shape[1]):
            for idx_z in range(0, kernel.shape[2]):
                current_weight = kernel[idx_x][idx_y][idx_z]

                if current_weight <= 0.0:
                    continue

                offsets.append(
                    [
                        idx_x - (kernel.shape[0] // 2),
                        idx_y - (kernel.shape[1] // 2),
                        idx_z
                    ]
                )

                weights.append(current_weight)

    return kernel, np.array(weights, dtype="float32"), np.array(offsets, dtype=int)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_max(values, weights):
    "Get the weighted maximum"
    idx = np.argmax(np.multiply(values, weights))
    return values[idx]


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_min(values, weights):
    "Get the weighted minimum"
    max_val = values.max()
    adjusted_values = np.where(weights == 0.0, max_val, values)
    idx = np.argmin(np.divide(adjusted_values, weights + 1e-7))
    return values[idx]


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_sum(values, weights):
    "Get the weighted sum"
    return np.sum(np.multiply(values, weights))


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_quantile(values, weights, quant):
    "Get the weighted median"
    sort_mask = np.argsort(values)
    sorted_data = values[sort_mask]
    sorted_weights = weights[sort_mask]
    cumsum = np.cumsum(sorted_weights)
    intersect = (cumsum - 0.5 * sorted_weights) / cumsum[-1]
    return np.interp(quant, intersect, sorted_data)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def hood_standard_deviation(values, weights):
    "Get the weighted standard deviation"
    summed = hood_sum(values, weights)
    variance = np.sum(np.multiply(np.power(np.subtract(values, summed), 2), weights))
    return np.sqrt(variance)


# @jit(nopython=True, parallel=True, nogil=True)
def convolve_array(arr, offsets, weights, method="sum", nodata=False, nodata_value=-9999.9, pad=False, pad_size=1, pad_type="edge"):
    """ Convolve an image with a function. """

    if pad:
        arr = np.pad(arr, pad_size, mode=pad_type)

    x_adj = arr.shape[0] - 1
    y_adj = arr.shape[1] - 1
    z_adj = (arr.shape[2] - 1) // 2

    hood_size = len(offsets)

    result = np.zeros((arr.shape[0], arr.shape[1], 1), dtype="float32")

    for idx_x in prange(arr.shape[0]):
        for idx_y in prange(arr.shape[1]):

            hood_values = np.zeros(hood_size, dtype="float32")
            hood_weights = np.zeros(hood_size, dtype="float32")
            weight_sum = np.array([0.0], dtype="float32")
            normalise = False

            for idx_n in range(hood_size):
                offset_x = idx_x + offsets[idx_n][0]
                offset_y = idx_y + offsets[idx_n][1]
                offset_z = offsets[idx_n][2]

                outside = False

                if offset_z < -z_adj:
                    offset_z = -z_adj
                    outside = True
                elif offset_z > z_adj:
                    offset_z = z_adj
                    outside = True

                if offset_x < 0:
                    offset_x = 0
                    outside = True
                elif offset_x > x_adj:
                    offset_x = x_adj
                    outside = True

                if offset_y < 0:
                    offset_y = 0
                    outside = True
                elif offset_y > y_adj:
                    offset_y = y_adj
                    outside = True

                value = arr[offset_x, offset_y, offset_z]

                if outside is True:
                    normalise = True
                    hood_weights[idx_n] = 0
                elif nodata and value == nodata_value:
                    normalise = True
                    hood_weights[idx_n] = 0
                else:
                    hood_values[idx_n] = value
                    weight = weights[idx_n]

                    hood_weights[idx_n] = weight
                    weight_sum[0] += weight

            if normalise and weight_sum[0] > 0.0:
                hood_weights = np.divide(hood_weights, weight_sum[0])

            if method == "sum":
                result[idx_x, idx_y, 0] = hood_sum(hood_values, hood_weights)
            elif method == "max":
                result[idx_x, idx_y, 0] = hood_max(hood_values, hood_weights)
            elif method == "min":
                result[idx_x, idx_y, 0] = hood_min(hood_values, hood_weights)
            elif method == "median":
                result[idx_x, idx_y, 0] = hood_quantile(hood_values, hood_weights, 0.5)
            elif method == "std":
                result[idx_x, idx_y, 0] = hood_standard_deviation(hood_values, hood_weights)
            else:
                raise ValueError("Unknown method passed to convolver")

    if pad:
        return result[
            pad_size:-pad_size,
            pad_size:-pad_size,
            :,
        ]

    return result



"""
### Utility functions to work with GDAL ###

These functions are used to interact with basic GDAL objects.

TODO:
    * Should file_path_lists be able to handle mixed inputs?
    * Make delete_raster_or_vector and the like accept a list of file paths?
"""

# Standard Library
import sys; sys.path.append("../../")
from pathlib import PurePosixPath
import os

# External
import numpy as np
from osgeo import gdal, ogr, osr

# Internal
from buteo.utils import core_utils, gdal_enums



def default_creation_options(options=None):
    """
    Takes a list of GDAL creation options and adds the following defaults to it if their not specified: </br>

    * `"TILED=YES"`
    * `"NUM_THREADS=ALL_CPUS"`
    * `"BIGG_TIF=YES"`
    * `"COMPRESS=LZW"`

    If any of the options are already specified, they are not added.

    ## Kwargs:
    `options` (_list_/None): The GDAL creation options to add to. (Default: **None**)

    ## Returns:
    (_list_): A list containing the default values.
    """
    assert isinstance(options, (list, type(None))), "Options must be a list or None."

    if options is None:
        options = []

    internal_options = list(options)

    opt_str = " ".join(internal_options)
    if "TILED" not in opt_str:
        internal_options.append("TILED=YES")

    if "NUM_THREADS" not in opt_str:
        internal_options.append("NUM_THREADS=ALL_CPUS")

    if "BIGTIFF" not in opt_str:
        internal_options.append("BIGTIFF=YES")

    if "COMPRESS" not in opt_str:
        internal_options.append("COMPRESS=LZW")

    return internal_options


def is_valid_datatype(file_path):
    """
    Check if a file path has a valid GDAL or OGR driver.

    ## Args:
    `file_path` (_str_): The file path to check.

    ## Returns:
    (_bool_): **True** if the file path is a valid GDAL or OGR driver, **False** otherwise.
    """
    assert isinstance(file_path, str), "file_path must be a string."

    ext = core_utils.path_to_ext(file_path)

    if ext in gdal_enums.get_valid_driver_extensions():
        return True

    return False


def is_valid_raster_datatype(file_path):
    """
    Check if a file path has a valid GDAL driver.

    ## Args:
    `file_path` (_str_): The file path to check.

    ## Returns:
    (_bool_): **True** if the file path is a valid GDAL Raster driver, **False** otherwise.
    """
    assert isinstance(file_path, str), "file_path must be a string."

    ext = core_utils.path_to_ext(file_path)

    if ext in gdal_enums.get_valid_raster_driver_extensions():
        return True

    return False


def is_valid_vector_datatype(file_path):
    """
    Check if a file path has a valid OGR driver.

    ## Args:
    `file_path` (_str_): The file path to check.

    ## Returns:
    (_bool_): **True** if the file path is a valid OGR Vector driver, **False** otherwise.
    """
    assert isinstance(file_path, str), "file_path must be a string."

    ext = core_utils.path_to_ext(file_path)

    if ext in gdal_enums.get_valid_vector_driver_extensions():
        return True

    return False


def path_to_driver(file_path):
    """
    Convert a file path to a GDAL or OGR driver ShortName (e.g. "GTiff" for "new_york.tif")

    ## Args:
    `file_path` (_str_): The file path to convert.

    ## Returns:
    (_str_): The GDAL or OGR driver ShortName.

    ## Raises:
    `ValueError`: If the file path is not a valid GDAL or OGR driver.
    """
    assert isinstance(file_path, str), "file_path must be a string."

    ext = core_utils.path_to_ext(file_path)

    if is_valid_datatype(ext):
        return gdal_enums.convert_extension_to_driver_shortname(ext)

    raise ValueError(f"Unable to parse GDAL or OGR driver from path: {file_path}")


def path_to_driver_vector(file_path):
    """
    Convert a file path to an OGR driver ShortName (e.g. "FlatGeoBuf" for "new_york.fgb")

    ## Args:
    `file_path` (_str_): The file path to convert.

    ## Returns:
    (_str_): The OGR driver ShortName.

    ## Raises:
    `ValueError`: If the file path is not a valid OGR driver.
    """
    assert isinstance(file_path, str), "file_path must be a string."

    ext = core_utils.path_to_ext(file_path)

    if is_valid_vector_datatype(file_path):
        return gdal_enums.convert_vector_extension_to_driver_shortname(ext)

    raise ValueError(f"Unable to parse GDAL or OGR driver from path: {file_path}")


def path_to_driver_raster(file_path):
    """
    Convert a file path to a GDAL driver ShortName (e.g. "GTiff" for "new_york.tif")

    ## Args:
    `file_path` (_str_): The file path to convert.

    ## Returns:
    (_str_): The GDAL driver ShortName.

    ## Raises:
    `ValueError`: If the file path is not a valid GDAL driver.
    """
    assert isinstance(file_path, str), "file_path must be a string."

    ext = core_utils.path_to_ext(file_path)

    if is_valid_raster_datatype(file_path):
        return gdal_enums.convert_raster_extension_to_driver_shortname(ext)

    raise ValueError(f"Unable to parse GDAL or OGR driver from path: {file_path}")


def is_in_memory(raster_or_vector):
    """
    Check if vector is in memory

    ## Args:
    `raster_or_vector` (_str_/_gdal.Dataset_/ogr.DataSource): The vector or raster to check.

    ## Returns:
    (_bool_): **True** if the vector is in memory, **False** otherwise.
    """
    assert isinstance(raster_or_vector, (str, gdal.Dataset, ogr.DataSource)), "raster_or_vector must be a string, gdal.Dataset, or ogr.DataSource."

    if isinstance(raster_or_vector, str):
        if core_utils.is_valid_mem_path(raster_or_vector):
            return True

        return False

    elif isinstance(raster_or_vector, (gdal.Dataset, ogr.DataSource)):
        driver = raster_or_vector.GetDriver()
        driver_short_name = None
        try:
            driver_short_name = driver.GetName()
        except Exception:
            driver_short_name = driver.ShortName

        if driver_short_name== "MEM":
            return True

        if driver_short_name == "Memory":
            return True

        if driver_short_name == "VirtualMem":
            return True

        if driver_short_name == "VirtualOGR":
            return True

        if core_utils.is_valid_mem_path(raster_or_vector.GetDescription()):
            return True

        return False

    else:
        raise TypeError("vector_or_raster must be a string, ogr.DataSource, or gdal.Dataset")


def delete_if_in_memory(raster_or_vector):
    """
    Delete raster or vector if it is in memory

    ## Args:
    `raster_or_vector` (_str_/_gdal.Dataset_/_ogr.DataSource_): The vector or raster to check.

    ## Returns:
    (_bool_): **True** if the vector is deleted, **False** otherwise.
    """
    if not isinstance(raster_or_vector, (str, gdal.Dataset, ogr.DataSource)):
        return False

    path = get_path_from_dataset(raster_or_vector)

    if is_in_memory(raster_or_vector):
        if isinstance(raster_or_vector, str):
            gdal.Unlink(raster_or_vector)
        else:
            raster_or_vector.Destroy()
            raster_or_vector = None

            gdal.Unlink(path)

        datasets = [ds.name for ds in gdal.listdir("/vsimem")]
        if path not in datasets:
            return True

    return False


def delete_if_in_memory_list(list_of_raster_or_vectors):
    """
    Deletes a list of raster or vector if they are in memory

    ## Args:
    `list_of_raster_or_vectors` (_list_): The vector(s) or raster(s) to check.

    ## Returns:
    (_bool_): **True** if they were all deleted, **False** otherwise.
    """
    assert isinstance(list_of_raster_or_vectors, list), "list_of_raster_or_vectors must be a list."

    deleted = []
    for raster_or_vector in list_of_raster_or_vectors:
        deleted.append(delete_if_in_memory(raster_or_vector))

    if any(deleted):
        return True

    return False


def delete_raster_or_vector(raster_or_vector):
    """
    Delete raster or vector. Can be used on both in memory and on disk.

    ## Args:
    `raster_or_vector` (_str_/_gdal.Dataset_/_ogr.DataSource_): The vector or raster to check.

    ## Returns:
    (_bool_): **True** if the file is deleted, **False** otherwise.
    """
    assert isinstance(raster_or_vector, (str, gdal.Dataset, ogr.DataSource)), "raster_or_vector must be a string, gdal.Dataset, or ogr.DataSource."

    if delete_if_in_memory(raster_or_vector):
        return True

    driver_shortname = path_to_driver(raster_or_vector)
    driver = gdal.GetDriverByName(driver_shortname)
    driver.Delete(raster_or_vector)

    if not core_utils.file_exists(raster_or_vector):
        return True

    return core_utils.delete_file(raster_or_vector)


def is_raster_empty(raster):
    """
    Check if a raster has bands or zero width and zero height.

    ## Args:
    `raster` (_gdal.Dataset_): The raster to check.

    ## Returns:
    (_bool_): **True** if the raster has bands, **False** otherwise.
    """
    assert isinstance(raster, gdal.Dataset), "raster must be a gdal.Dataset."

    if raster.RasterCount == 0:
        return True

    if raster.RasterXSize == 0 or raster.RasterYSize == 0:
        return True

    return False


def is_vector_empty(vector):
    """
    Check if a vector has features with geometries

    ## Args:
    `vector` (_ogr.DataSource_): The vector to check.

    ## Returns:
    (_bool_): **True** if the vector has features, **False** otherwise.
    """
    assert isinstance(vector, ogr.DataSource), "vector must be an ogr.DataSource."

    layer_count = vector.GetLayerCount()

    if layer_count == 0:
        return True

    for layer in range(0, layer_count):
        layer = vector.GetLayerByIndex(layer)

        if layer.GetFeatureCount() > 0:
            feature_count = layer.GetFeatureCount()

            for feature in range(0, feature_count):
                feature = layer.GetNextFeature()

                if feature.GetGeometryRef() is not None:
                    return False

    return True


def get_gdal_memory():
    """ Get at list of all active memory layers in GDAL. """
    datasets = [ds.name for ds in gdal.listdir("/vsimem")]
    return datasets


def clear_gdal_memory():
    """ Clears all gdal memory. """
    memory = get_gdal_memory()

    for dataset in memory:
        gdal.Unlink(dataset)

    if len(get_gdal_memory()) != 0:
        for dataset in get_gdal_memory():
            opened = None
            mem_path = PurePosixPath("/vsimem", dataset).as_posix()
            if is_raster(mem_path):
                opened = gdal.Open(mem_path)
            elif is_raster(mem_path):
                opened = ogr.Open(mem_path)
            else:
                raise ValueError(f"Unable to open dataset: {mem_path}")
            driver = opened.GetDriver()
            driver.Delete(dataset)

            opened = None
            driver = None

    if len(get_gdal_memory()) != 0:
        print("Failed to clear all GDAL memory.")


def gdal_print_memory():
    """ Prints all gdal memory. """
    memory = get_gdal_memory()

    for dataset in memory:
        print(dataset)


def is_raster(potential_raster, *, empty_is_invalid=True):
    """Checks if a variable is a valid raster.

    ## Args:
    `potential_raster` (_any_): The variable to check.

    ## Kwargs:
    `empty_is_invalid` (_bool_): If **True**, an empty raster is considered invalid. (Default: **True**)

    ## Returns:
    (_bool_): **True** if the variable is a valid raster, **False** otherwise.
    """
    if isinstance(potential_raster, str):
        if not core_utils.file_exists(potential_raster) and not core_utils.is_valid_mem_path(potential_raster):
            return False

        try:
            gdal.PushErrorHandler('CPLQuietErrorHandler')
            opened = gdal.Open(potential_raster, 0)
            gdal.PopErrorHandler()
        except Exception:
            return False

        if opened is None:
            return False

        if empty_is_invalid and is_raster_empty(opened):
            return False

        opened = None

        return True

    if isinstance(potential_raster, gdal.Dataset):

        if empty_is_invalid and is_raster_empty(potential_raster):
            return False

        return True

    return False


def is_raster_list(potential_raster_list, *, empty_is_invalid=True):
    """
    Checks if a variable is a valid list of rasters.

    ## Args:
    `potential_raster_list` (_any_): The variable to check.

    ## Kwargs:
    `empty_is_invalid` (_bool_): If **True**, an empty raster is considered invalid. (Default: **True**)
    """
    if not isinstance(potential_raster_list, list):
        return False

    if len(potential_raster_list) == 0:
        return False

    for element in potential_raster_list:
        if not is_raster(element, empty_is_invalid=empty_is_invalid):
            return False

    return True


def is_vector(potential_vector, empty_is_invalid=True):
    """
    Checks if a variable is a valid vector.

    ## Args:
    `potential_vector` (_any_): The variable to check.

    ## Kwargs:
    `empty_is_invalid` (_bool_): If **True**, an empty vector is considered invalid. (Default: **True**)

    ## Returns:
    (_bool_): **True** if the variable is a valid vector, **False** otherwise.
    """
    if isinstance(potential_vector, ogr.DataSource):

        if empty_is_invalid and is_vector_empty(potential_vector):
            return False

        return True

    if isinstance(potential_vector, str):
        gdal.PushErrorHandler("CPLQuietErrorHandler")

        opened = ogr.Open(potential_vector, 0)
        if opened is None:
            extension = os.path.splitext(potential_vector)[1][1:]

            if extension == "memory" or "mem":
                driver = ogr.GetDriverByName("Memory")
                opened = driver.Open(potential_vector)

        gdal.PopErrorHandler()

        if isinstance(opened, ogr.DataSource):

            if empty_is_invalid and is_vector_empty(opened):
                return False

            return True

    return False


def is_vector_list(potential_vector_list, *, empty_is_invalid=True):
    """
    Checks if a variable is a valid list of vectors.

    ## Args:
    `potential_vector_list` (_any_): The variable to check.

    ## Kwargs:
    `empty_is_invalid` (_bool_): If **True**, an empty vector is considered invalid. (Default: **True**)
    """
    if not isinstance(potential_vector_list, list):
        return False

    if len(potential_vector_list) == 0:
        return False

    for element in potential_vector_list:
        if not is_vector(element, empty_is_invalid=empty_is_invalid):
            return False

    return True


def is_raster_or_vector(potential_raster_or_vector, *, empty_is_invalid=True):
    """
    Checks if a variable is a valid raster or vector.

    ## Args:
    `potential_raster_or_vector` (_any_): The variable to check.

    ## Kwargs:
    `empty_is_invalid` (_bool_): If **True**, an empty raster or vector is considered invalid. (Default: **True**)
    """
    if is_raster(potential_raster_or_vector, empty_is_invalid=empty_is_invalid):
        return True

    if is_vector(potential_raster_or_vector, empty_is_invalid=empty_is_invalid):
        return True

    return False


def is_raster_or_vector_list(potential_raster_or_vector_list, *, empty_is_invalid=True):
    """
    Checks if a variable is a valid list of rasters or vectors.

    ## Args:
    `potential_raster_or_vector_list` (_any_): The variable to check.

    ## Kwargs:
    `empty_is_invalid` (_bool_): If **True**, an empty raster or vector is considered invalid. (Default: **True**)
    """
    if not isinstance(potential_raster_or_vector_list, list):
        return False

    if len(potential_raster_or_vector_list) == 0:
        return False

    for element in potential_raster_or_vector_list:
        if not is_raster_or_vector(element, empty_is_invalid=empty_is_invalid):
            return False

    return True


def create_memory_path(path, *, prefix="", suffix="", add_uuid=True):
    """
    Gets a memory path from a string in the format: </br>
    `/vsimem/prefix_basename_time_uuid_suffix.ext`

    ## Args:
    `path` (_str_): The path to the original file. </br>

    ## Kwargs:
    `prefix` (_str_): The prefix to add to the memory path. (**Default**: `""`) </br>
    `suffix` (_str_): The suffix to add to the memory path. (**Default**: `""`) </br>
    `add_uuid` (_bool_): If True, add a uuid to the memory path. (**Default**: `True`) </br>

    ## Returns:
    (_str_): A string to the memory path. `/vsimem/prefix_basename_time_uuid_suffix.ext`
    """
    assert isinstance(path, str), f"path must be a string. Received: {path}"
    assert len(path) > 0, f"path must not be empty. Received: {path}"
    assert not core_utils.is_valid_mem_path(path), f"path must not be a valid memory path. Received: {path}"

    basename = os.path.basename(path)

    return core_utils.get_augmented_path(
        PurePosixPath("/vsimem", basename).as_posix(),
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        folder=None,
    )


def get_path_from_dataset(dataset, *, dataset_type=None):
    """
    Gets the path from a datasets. Can be vector or raster, string or opened.

    ## Args:
    `dataset` (_str_/_gdal.Dataset_/_ogr.DataSource_): The dataset.

    ## Kwargs:
    `dataset_type` (_str_): The type of dataset. If not specified, it is guessed. (**Default**: `None`)

    ## Returns:
    (_list_): The path to the dataset.
    """
    if (dataset_type == "raster" or dataset_type is None) and is_raster(dataset, empty_is_invalid=False):
        if isinstance(dataset, str):
            raster = gdal.Open(dataset, 0)
        elif isinstance(dataset, gdal.Dataset):
            raster = dataset
        else:
            raise Exception(f"Could not read input raster: {raster}")

        path = raster.GetDescription()
        raster = None

        return path

    if (dataset_type == "vector" or dataset_type is None) and is_vector(dataset, empty_is_invalid=False):
        if isinstance(dataset, str):
            vector = ogr.Open(dataset, 0)
        elif isinstance(dataset, ogr.DataSource):
            vector = dataset
        else:
            raise Exception(f"Could not read input vector: {vector}")

        path = vector.GetDescription()
        vector = None

        return path

    raise ValueError("The dataset is not a raster or vector.")


def get_path_from_dataset_list(datasets, *, allow_mixed=False, dataset_type=None):
    """
    Gets the paths from a list of datasets.

    ## Args:
    `datasets` (_list_): The list of datasets.

    ## Kwargs:
    `allow_mixed` (_bool_): If True, allow mixed raster/vector datasets. (**Default**: `False`) </br>
    `dataset_type` (_str_/_None_): The type of dataset. If not specified, it is guessed. (**Default**: `None`)

    ## Returns:
    (_list_): A list of paths.
    """
    assert isinstance(datasets, list), "The datasets must be a list."
    assert isinstance(dataset_type, (str, type(None))), "The dataset_type must be 'raster', 'vector', or None."

    rasters = False
    vectors = False

    outputs = []
    for dataset in datasets:
        if (dataset_type == "raster" or dataset_type is None) and is_raster(dataset, empty_is_invalid=False):
            dataset_type = "raster"
            rasters = True
        elif (dataset_type == "vector" or dataset_type is None) and is_vector(dataset, empty_is_invalid=False):
            dataset_type = "vector"
            vectors = True
        else:
            raise ValueError(f"The dataset is not a raster or vector. {dataset}")

        if rasters and vectors and not allow_mixed:
            raise ValueError("The datasets cannot be vectors and rasters mixed.")

        outputs.append(get_path_from_dataset(dataset, dataset_type=dataset_type))

    return outputs


def convert_geom_to_vector(geom):
    """
    Converts a geometry to a vector.

    ## Args:
    `geom` (_ogr.Geometry_): The geometry to convert.

    ## Kwargs:
    `layer_name` (_str_): The name of the layer. (Default: **"geom"**)
    `add_uuid` (_bool_): If **True**, a UUID will be added to the layer. (Default: **True**)

    ## Returns:
    (_ogr.DataSource_): The vector.
    """
    assert isinstance(geom, ogr.Geometry), "geom must be an ogr.Geometry."

    path = create_memory_path("converted_geom.fgb", add_uuid=True)

    driver = ogr.GetDriverByName(path_to_driver_vector(path))
    vector = driver.CreateDataSource(path)

    layer = vector.CreateLayer("converted_geom", geom.GetSpatialReference(), geom.GetGeometryType())

    feature = ogr.Feature(layer.GetLayerDefn())
    feature.SetGeometry(geom)

    layer.CreateFeature(feature)
    feature.Destroy()

    return vector


def parse_projection(projection, *, return_wkt=False):
    """
    Parses a gdal, ogr og osr data source and extraction the projection. If
    a string or int is passed, it attempts to open it and return the projection as
    an osr.SpatialReference.

    ## Args:
    `projection` (_str_/_int_/_gdal.Dataset_/_ogr.DataSource_/_osr.SpatialReference_): The projection to parse.

    ## Kwargs:
    `return_wkt` (_bool_): If **True** the projection will be returned as a WKT string, otherwise an osr.SpatialReference is returned. (Default: **False**)
    """
    assert isinstance(projection, (str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference)), "projection must be a string, int, gdal.Dataset, ogr.DataSource, or osr.SpatialReference."

    err_msg = f"Unable to parse target projection: {projection}"
    target_proj = osr.SpatialReference()

    # Suppress gdal errors and handle them ourselves.
    gdal.PushErrorHandler("CPLQuietErrorHandler")

    if isinstance(projection, ogr.DataSource):
        layer = projection.GetLayer()
        target_proj = layer.GetSpatialRef()

    elif isinstance(projection, gdal.Dataset):
        target_proj.ImportFromWkt(projection.GetProjection())

    elif isinstance(projection, osr.SpatialReference):
        target_proj = projection

    elif isinstance(projection, str):
        if is_raster(projection):
            ref = gdal.Open(projection, 0)
            target_proj.ImportFromWkt(ref.GetProjection())
        elif is_vector(projection):
            ref = ogr.Open(projection, 0)
            layer = ref.GetLayer()
            target_proj = layer.GetSpatialRef()
        else:
            code = target_proj.ImportFromWkt(projection)
            if code != 0:
                code = target_proj.ImportFromProj4(projection)
                if code != 0:
                    raise ValueError(err_msg)

    elif isinstance(projection, int):
        code = target_proj.ImportFromEPSG(projection)
        if code != 0:
            raise ValueError(err_msg)

    else:
        raise ValueError(err_msg)

    gdal.PopErrorHandler()

    if isinstance(target_proj, osr.SpatialReference):
        if target_proj.GetName() is None:
            raise ValueError(err_msg)

        if return_wkt:
            return target_proj.ExportToWkt()

        return target_proj
    else:
        raise ValueError(err_msg)


def projections_match(source1, source2):
    """
    Tests if two projection sources have the same projection.

    ## Args:
    `source1` (_str_/_int_/_gdal.Dataset_/_ogr.DataSource_/_osr.SpatialReference_): The first projection to parse.
    `source1` (_str_/_int_/_gdal.Dataset_/_ogr.DataSource_/_osr.SpatialReference_): The second projection to parse.

    ## Returns:
    (_bool_): **True** if the projections match, **False** otherwise.
    """
    proj1 = parse_projection(source1)
    proj2 = parse_projection(source2)

    if proj1.IsSame(proj2):
        return True
    elif proj1.ExportToProj4() == proj2.ExportToProj4():
        return True

    return False


def projections_match_list(list_of_projection_sources):
    """
    Tests if a list of projection sources all have the same projection.

    ## Args:
    `list_of_projection_sources` (_list_): A list of projections to test.

    ## Returns:
    (_bool_): **True** if the projections match, **False** otherwise.
    """
    assert isinstance(list_of_projection_sources, list), "list_of_projection_sources must be a list."
    assert len(list_of_projection_sources) > 1, "The list 'list_of_projection_sources' must be a list of len > 1"

    first = None
    for index, source in enumerate(list_of_projection_sources):
        if index == 0:
            first = parse_projection(source)
        else:
            compare = parse_projection(source)

            if not first.IsSame(compare) and first.ExportToProj4() != compare.ExportToProj4():
                return False

    return True


def parse_raster_size(target, *, target_in_pixels=False):
    """
    Parses the raster size from either a list of numbers or a GDAL raster.

    ## Args:
    `target_size` (_any_): The target to parse raster_size from.

    ## Kwargs:
    `target_in_pixels` (_bool_): If **True**, the target size is in pixels, otherwise it is in the rasters_units. (Default: **False**)

    ## Returns:
    (_tuple_): The raster size in the form of: `(x_res, y_res, x_size, y_size)`.
    """
    assert target is not None, "target_size cannot be None."

    x_res = None
    y_res = None

    x_pixels = None
    y_pixels = None

    if isinstance(target, (gdal.Dataset, str)):
        reference = (
            target
            if isinstance(target, gdal.Dataset)
            else gdal.Open(target, 0)
        )

        transform = reference.GetGeoTransform()
        reference = None

        x_res = transform[1]
        y_res = abs(transform[5])

    elif target_in_pixels:
        if isinstance(target, tuple) or isinstance(target, list):
            if len(target) == 1:
                if core_utils.is_number(target[0]):
                    x_pixels = int(target[0])
                    y_pixels = int(target[0])
                else:
                    raise ValueError(
                        "target_size_pixels is not a number or a list/tuple of numbers."
                    )
            elif len(target) == 2:
                if core_utils.is_number(target[0]) and core_utils.is_number(target[1]):
                    x_pixels = int(target[0])
                    y_pixels = int(target[1])
            else:
                raise ValueError("target_size_pixels is either empty or larger than 2.")
        elif core_utils.is_number(target):
            x_pixels = int(target)
            y_pixels = int(target)
        else:
            raise ValueError("target_size_pixels is invalid.")

        x_res = None
        y_res = None
    else:
        if isinstance(target, tuple) or isinstance(target, list):
            if len(target) == 1:
                if core_utils.is_number(target[0]):
                    x_res = float(target[0])
                    y_res = float(target[0])
                else:
                    raise ValueError(
                        "target_size is not a number or a list/tuple of numbers."
                    )
            elif len(target) == 2:
                if core_utils.is_number(target[0]) and core_utils.is_number(target[1]):
                    x_res = float(target[0])
                    y_res = float(target[1])
            else:
                raise ValueError("target_size is either empty or larger than 2.")
        elif core_utils.is_number(target):
            x_res = float(target)
            y_res = float(target)
        else:
            raise ValueError("target_size is invalid.")

        x_pixels = None
        y_pixels = None

    return x_res, y_res, x_pixels, y_pixels


def get_gdalwarp_ram_limit(limit_in_mb):
    """
    Converts a RAM limit to a GDALWarp RAM limit.

    ## Args:
    `limit` (_str_/_int_): The RAM limit to convert. Can be auto, a percentage "80%" or a number in MB.

    ## Returns:
    (_int_): The GDALWarp RAM limit in bytes.
    """
    assert isinstance(limit_in_mb, (str, int)), "limit must be a string or integer."

    min_ram = 1000000
    limit = min_ram

    if isinstance(limit_in_mb, str):
        if limit_in_mb.lower() == "auto":
            return core_utils.get_dynamic_memory_limit_bytes()
        else:
            if "%" not in limit_in_mb:
                raise ValueError(f"Invalid limit: {limit_in_mb}")

            limit_in_percentage = limit_in_mb.replace("%", "")
            limit_in_percentage = int(limit_in_percentage)

            if limit_in_percentage <= 0 or limit_in_percentage > 100:
                raise ValueError(f"Invalid limit: {limit_in_mb}")

            limit = core_utils.get_percentage_of_total_ram_mb(limit_in_percentage) * (1024 ** 2)

            if limit > min_ram:
                return limit

    if limit > min_ram:
        return int(limit_in_mb * (1024 ** 2))

    return min_ram


def to_array_list(array_or_list_of_array):
    """
    Converts a numpy array or list of numpy arrays to a list of arrays.

    ## Args:
    `array_or_list_of_array` (_numpy.ndarray_/_list_/_str_): The numpy array or list of numpy arrays to convert to a list of arrays.

    ## Returns:
    (_list_): The list of arrays.
    """
    assert isinstance(array_or_list_of_array, (np.ndarray, list, str)), "array_or_list_of_array must be a numpy array, list of numpy arrays, or string."

    return_list = [array_or_list_of_array] if isinstance(array_or_list_of_array, np.ndarray) else array_or_list_of_array

    if len(return_list) == 0:
        raise ValueError("Empty array list.")

    for array in return_list:
        if not isinstance(array, np.ndarray):
            if isinstance(array, str) and core_utils.file_exists(array):
                try:
                    _ = np.load(array)
                except:
                    raise ValueError(f"Invalid array in list: {array}") from None
        else:
            raise ValueError(f"Invalid array in list: {array}")

    return return_list


def to_band_list(
    band_number,
    band_count,
):
    """
    Converts a band number or list of band numbers to a list of band numbers.

    ## Args:
    `band_number` (_int_/_float_/_list_): The band number or list of band numbers to convert to a list of band numbers. </br>
    `band_count` (_int_): The number of bands in the raster. </br>

    ## Returns:
    (_list_): The list of band numbers.
    """

    return_list = []
    if not isinstance(band_number, (int, float, list)):
        raise TypeError(f"Invalid type for band: {type(band_number)}")

    if isinstance(band_number, list):
        if len(band_number) == 0:
            raise ValueError("Provided list of bands is empty.")
        for val in band_number:
            try:
                band_int = int(val)
            except Exception:
                raise ValueError(
                    f"List of bands contained non-valid band number: {val}"
                ) from None

            if band_int > band_count - 1:
                raise ValueError("Requested a higher band that is available in raster.")
            else:
                return_list.append(band_int)
    elif band_number == -1:
        for val in range(band_count):
            return_list.append(val)
    else:
        if band_number > band_count + 1:
            raise ValueError("Requested a higher band that is available in raster.")
        else:
            return_list.append(int(band_number))

    return return_list


def create_output_path(
    dataset_path,
    out_path=None,
    *,
    overwrite=True,
    prefix="",
    suffix="",
    add_uuid=False,
):
    """
    Prepares a raster/vector for writing. Generates an output path. If no output path is
    specified, the raster is written to memory. If a folder is given, that output directory is used
    along with the input filename.

    ## Args:
    `dataset_path` (_gdal.Dataset_/_ogr.DataSource_/_str_): A **GDAL** or **OGR** dataframe, a name (with extension) of a raster </br>

    ## Kwargs:
    `out_path` (_str_/_None_): A path to a directory to write the raster to. (Default: **None**). </br>
    `overwrite` (_bool_): If True, the output raster will be overwritten if it already exists. (Default: **True**). </br>
    `prefix` (_str_): A string to prepend to the output filename. (Default: **""**). </br>
    `suffix` (_str_): A string to append to the output filename. (Default: **""**). </br>
    `add_uuid` (_bool_): If True, a UUID will be added to the output filename. (Default: **False**). </br>

    ## Returns:
    (_str_): A path to the output raster or a list of paths.
    """
    assert isinstance(dataset_path, (str)), "dataset_path must be a string or a list of strings."
    assert len(dataset_path) > 0, "dataset_path must be a path of len larger than 0."
    assert isinstance(out_path, (str, type(None))), "out_path must be a string or None."

    if out_path is not None:
        assert core_utils.is_valid_output_path(out_path, overwrite=overwrite), "out_path must be a valid output path or None."

    aug_path = None
    if out_path is None:

        # Not all drivers are support in memory. So if nothing is specified,
        # we'll convert to tif and fgb for memory files.

        og_path = get_path_from_dataset(dataset_path)
        og_ext = core_utils.path_to_ext(og_path)

        if gdal_enums.is_valid_raster_driver_extension(og_ext):
            og_path = os.path.basename(core_utils.change_path_ext(og_path, "tif"))

        elif gdal_enums.is_valid_vector_driver_extension(og_ext):
            og_path = os.path.basename(core_utils.change_path_ext(og_path, "fgb"))

        aug_path = core_utils.get_augmented_path(
            og_path,
            prefix=prefix,
            suffix=suffix,
            add_uuid=add_uuid,
            folder="/vsimem",
        )
    elif core_utils.folder_exists(core_utils.path_to_folder(out_path)):
        aug_path = core_utils.get_augmented_path(
            os.path.basename(dataset_path),
            prefix=prefix,
            suffix=suffix,
            add_uuid=add_uuid,
            folder=core_utils.path_to_folder(out_path),
        )
    elif core_utils.is_valid_mem_path(out_path):
        aug_path = core_utils.get_augmented_path(
            os.path.basename(dataset_path),
            prefix=prefix,
            suffix=suffix,
            add_uuid=add_uuid,
            folder="/vsimem",
        )
    else:
        raise ValueError("out_path must be a valid path or a list of valid paths.")

    assert core_utils.is_valid_output_path(aug_path, overwrite=overwrite), f"Error while generating output. It is invalid: {aug_path}"

    return aug_path


def create_output_path_list(
    dataset_path,
    out_path=None,
    *,
    overwrite=True,
    prefix="",
    suffix="",
    add_uuid=False,
):
    """
    Prepares a raster/vector or a list of rasters/vectors for writing. Generates output paths. If no output paths are
    specified, the rasters are written to memory. If a folder is given, the output directory is chosen,
    the input filenames remain the same. If a specific path is used it must be the same length as the
    input.

    ## Args:
    `dataset_path` (_gdal.Dataset_/_ogr.DataSource_/_str_/_list__): A **GDAL** or **OGR** dataframe, a path to a raster or a list of same. </br>

    ## Kwargs:
    `out_path` (_str_/_None_): A path to a directory to write the raster to. (Default: **None**). </br>
    `overwrite` (_bool_): If True, the output raster will be overwritten if it already exists. (Default: **True**). </br>
    `prefix` (_str_): A string to prepend to the output filename. (Default: **""**). </br>
    `suffix` (_str_): A string to append to the output filename. (Default: **""**). </br>
    `add_uuid` (_bool_): If True, a UUID will be added to the output filename. (Default: **False**). </br>

    ## Returns:
    (_str_/_list_): A path to the output raster or a list of paths.
    """
    assert isinstance(dataset_path, (list)), "dataset_path must be a string or a list of strings."
    assert isinstance(out_path, (str, type(None))), "out_path must be a string or None."

    assert len(dataset_path) > 0, "dataset_path must contain at least one path."

    if isinstance(out_path, list):
        assert len(out_path) == len(dataset_path), "out_path must be the same length as dataset_path if a list is provided."
        assert core_utils.is_valid_output_path_list(out_path, overwrite=overwrite), "out_path must be a list of valid output paths."

    output = []
    out_path = core_utils.ensure_list(out_path)
    if len(out_path) != len(dataset_path):
        out_path = out_path * len(dataset_path)

    for index, path in enumerate(dataset_path):
        output.append(create_output_path(
            path,
            out_path=out_path[index],
            overwrite=overwrite,
            prefix=prefix,
            suffix=suffix,
            add_uuid=add_uuid,
        ))

    assert core_utils.is_valid_output_path_list(output, overwrite=overwrite), f"Error while generating outputs. They are invalid: {output}"

    return output


def save_dataset_to_disk(
    dataset,
    out_path,
    *,
    overwrite=True,
    creation_options=None,
    prefix="",
    suffix="",
    add_uuid=False,
):
    """
    Writes a dataset to disk. Can be a raster or a vector.

    ## Args:
    `dataset` (_str_/_gdal.Dataset_/_ogr.DataSource_/_list_): The dataset(s) to save. </br>
    `out_path` (_str_/_list_): The path(s) to save the dataset(s) to. </br>

    ## Kwargs:
    `overwrite` (_bool_): Whether to overwrite the file if it already exists. (Default: **True**) </br>
    `creation_options` (_list_/_None_): A list of creation options to pass to GDAL if saving as raster. (Default: **True**) </br>
    `prefix` (_str_): A prefix to add to the file name. (Default: **""**) </br>
    `suffix` (_str_): A suffix to add to the file name. (Default: **""**) </br>
    `add_uuid` (_bool_): Whether to add a UUID to the file name. (Default: **False**) </br>
    """
    datasets = core_utils.ensure_list(dataset)
    datasets_paths = get_path_from_dataset_list(datasets, allow_mixed=True)
    out_paths = create_output_path_list(datasets_paths, out_path, prefix=prefix, suffix=suffix, add_uuid=add_uuid)

    options = None

    for index, dataset_ in enumerate(datasets):
        opened_dataset = None
        dataset_type = None

        if is_raster(dataset_):
            options = default_creation_options(creation_options)
            dataset_type = "raster"
            if isinstance(dataset_, str):
                opened_dataset = gdal.Open(dataset_, 0)
            elif isinstance(dataset_, gdal.Dataset):
                opened_dataset = dataset_
            else:
                raise Exception(f"Could not read input raster: {dataset_}")

        elif is_vector(dataset_):
            dataset_type = "vector"
            if isinstance(dataset_, str):
                opened_dataset = ogr.Open(dataset_, 0)
            elif isinstance(dataset_, ogr.DataSource):
                opened_dataset = dataset_
            else:
                raise Exception(f"Could not read input vector: {dataset_}")

        else:
            raise Exception(f"Invalid dataset type: {dataset_}")

        driver_destination = None

        if dataset_type == "raster":
            driver_destination = gdal.GetDriverByName(path_to_driver_raster(out_paths[index]))
        else:
            driver_destination = ogr.GetDriverByName(path_to_driver_vector(out_paths[index]))

        assert driver_destination is not None, "Could not get driver for output dataset."

        core_utils.remove_if_required(out_paths[index], overwrite)

        driver_destination.CreateCopy(
            out_path[index],
            opened_dataset,
            options=options,
        )

    if isinstance(dataset, list):
        return out_paths[0]

    return out_paths


def save_dataset_to_memory(
    dataset,
    *,
    overwrite=True,
    creation_options=None,
    prefix="",
    suffix="",
    add_uuid=True,
):
    """
    Writes a dataset to memory. Can be a raster or a vector.

    ## Args:
    `dataset` (_str_/_gdal.Dataset_/_ogr.DataSource_/_list_): The dataset(s) to save. </br>

    ## Kwargs:
    `overwrite` (_bool_): Whether to overwrite the file if it already exists. (Default: **True**) </br>
    `creation_options` (_list_/_None_): A list of creation options to pass to GDAL if saving as raster. (Default: **True**) </br>
    `prefix` (_str_): A prefix to add to the file name. (Default: **""**) </br>
    `suffix` (_str_): A suffix to add to the file name. (Default: **""**) </br>
    `add_uuid` (_bool_): Whether to add a UUID to the file name. (Default: **True**) </br>
    """
    return save_dataset_to_disk(
        dataset,
        out_path=None,
        overwrite=overwrite,
        creation_options=creation_options,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )



"""
Calculate zonal statistics from vector and raster files.

BROKEN

TODO:
    - Fix rasterize step
    - Improve documentation

"""

import sys; sys.path.append("../../") # Path: buteo/vector/zonal_statistics.py
import numpy as np

from numba import jit
from osgeo import ogr

from buteo.raster.core_raster import raster_to_array, raster_to_metadata
from buteo.vector.rasterize import rasterize_vector
from buteo.vector.reproject import reproject_vector
from buteo.vector.core_vector import (
    open_vector,
    _vector_to_memory,
    _vector_to_metadata,
)
from buteo.vector.zonal_statistics_stats import calculate_array_stats
from buteo.utils.core_utils import progress



import numpy as np
from enum import Enum

from numba import jit


class stat(Enum):
    count = 1
    range = 2
    min = 3
    max = 4
    sum = 5
    mean = 6
    avg = 6
    average = 6
    var = 7
    variance = 7
    std = 8
    stdev = 8
    standard_deviation = 8
    skew = 9
    kurtosis = 10
    median = 11
    med = 11
    iqr = 12
    q02 = 13
    q98 = 14
    q1 = 15
    q3 = 16
    mad = 17
    median_absolute_deviation = 17
    mode = 18
    snr = 19
    eff = 20
    cv = 21


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def calculate_array_stats(arr, stats):
    stats_length = int(len(stats))
    result = np.zeros((stats_length), dtype="float32")

    for idx in range(stats_length):
        if stat(stats[idx]) == 1:  # count
            result[idx] = arr.size
        elif stat(stats[idx]) == 2:  # range
            result[idx] = np.ptp(arr)
        elif stat(stats[idx]) == 3:  # min
            result[idx] = np.min(arr)
        elif stat(stats[idx]) == 4:  # max
            result[idx] = np.max(arr)
        elif stat(stats[idx]) == 5:  # sum
            result[idx] = np.sum(arr)
        elif stat(stats[idx]) == 6:  # mean
            result[idx] = np.mean(arr)
        elif stat(stats[idx]) == 7:  # var
            result[idx] = np.var(arr)
        elif stat(stats[idx]) == 8:  # std
            result[idx] = np.std(arr)
        elif stat(stats[idx]) == 9:  # skew
            mean = np.mean(arr)
            std = np.std(arr)
            if std == 0:
                result[idx] = 0.0
                continue
            deviations = np.sum(np.power(arr - mean, 3))
            result[idx] = (deviations * (1 / arr.size)) / (np.power(std, 3))
        elif stat(stats[idx]) == 10:  # kurt
            mean = np.mean(arr)
            std = np.std(arr)
            if std == 0:
                result[idx] = 0.0
                continue
            deviations = np.sum(np.power(arr - mean, 4))
            result[idx] = (deviations * (1 / arr.size)) / (np.power(std, 4))
        elif stat(stats[idx]) == 11:  # median
            result[idx] = np.median(arr)
        elif stat(stats[idx]) == 12:  # iqr
            result[idx] = np.quantile(
                arr, np.array([0.25, 0.75], dtype="float32")
            ).sum()
        elif stat(stats[idx]) == 13:  # q02
            result[idx] = np.quantile(arr, 0.02)
        elif stat(stats[idx]) == 14:  # q92
            result[idx] = np.quantile(arr, 0.98)
        elif stat(stats[idx]) == 15:  # q1
            result[idx] = np.quantile(arr, 0.25)
        elif stat(stats[idx]) == 16:  # q3
            result[idx] = np.quantile(arr, 0.75)
        elif stat(stats[idx])== 17:  # mad
            median = np.median(arr)
            absdev = np.abs(arr - median)
            result[idx] = np.median(absdev)
        elif stat(stats[idx]) == 18:  # mode
            uniques = np.unique(arr)
            counts = np.zeros_like(uniques, dtype="uint64")

            # numba does not support return count of uniques.
            for idx_values in range(len(arr)):
                val = arr[idx_values]
                for idx_uniques in range(len(uniques)):
                    if val == uniques[idx_uniques]:
                        counts[idx_uniques] = counts[idx_uniques] + 1
            index = np.argmax(counts)
            result[idx] = uniques[index]
        elif stat(stats[idx]) == 19:  # snr
            std = np.std(arr)
            if std == 0:
                result[idx] = 0.0
                continue
            result[idx] = np.mean(arr) / std
        elif stat(stats[idx]) == 20:  # eff
            mean = np.mean(arr)
            if mean == 0:
                result[idx] = 0.0
                continue
            result[idx] = np.var(arr) / np.power(mean, 2)
        elif stat(stats[idx]) == 21:  # cv
            mean = np.mean(arr)
            if mean == 0:
                result[idx] = 0.0
                continue
            result[idx] = np.std(arr) / mean

    return result

@jit(nopython=True, nogil=True, fastmath=True, inline="always")
def overlap_size_calc(extent, raster_transform):
    return np.array(
        [
            (extent[1] - extent[0]) / raster_transform[1],
            (extent[3] - extent[2]) / abs(raster_transform[5]),
        ],
        dtype=np.int32,
    )


@jit(nopython=True, nogil=True, fastmath=True, inline="always")
def align_extent(raster_transform, vector_extent, raster_size):
    pixel_width = abs(raster_transform[1])
    pixel_height = abs(raster_transform[5])

    raster_min_x = raster_transform[0]
    # raster_max_x = raster_min_x + (raster_size[0] * pixel_width)
    raster_max_y = raster_transform[3]
    raster_min_y = raster_max_y + (raster_size[1] * -pixel_width)

    vector_min_x = vector_extent[0]
    vector_max_x = vector_extent[1]
    vector_min_y = vector_extent[2]
    vector_max_y = vector_extent[3]

    # Align the two extents
    vector_min_x = vector_min_x - (vector_min_x - raster_min_x) % pixel_width
    vector_max_x = vector_max_x + (vector_max_x - raster_min_x) % pixel_width
    vector_min_y = vector_min_y - (vector_min_y - raster_max_y) % pixel_height
    vector_max_y = vector_max_y + (vector_max_y - raster_max_y) % pixel_height

    rasterized_x_size = int((vector_max_x - vector_min_x) / pixel_width)
    rasterized_y_size = int((vector_max_y - vector_min_y) / pixel_height)

    rasterized_x_offset = int((vector_min_x - raster_min_x) / pixel_width)
    rasterized_y_offset = (
        int(raster_size[1] - int((vector_min_y - raster_min_y) / pixel_height))
        - rasterized_y_size
    )

    if rasterized_x_offset < 0:
        rasterized_x_offset = 0

    if rasterized_y_offset < 0:
        rasterized_y_offset = 0

    if (rasterized_x_offset + rasterized_x_size) > raster_size[0]:
        rasterized_x_offset = rasterized_x_offset - (
            (rasterized_x_offset + rasterized_x_size) - raster_size[0]
        )

    if (rasterized_y_offset + rasterized_y_size) > raster_size[1]:
        rasterized_y_offset = rasterized_y_offset - (
            (rasterized_y_offset + rasterized_y_size) - raster_size[1]
        )

    new_vector_extent = np.array(
        [vector_min_x, vector_max_x, vector_min_y, vector_max_y], dtype=np.float32
    )
    rasterized_size = np.array(
        [rasterized_x_size, rasterized_y_size, pixel_width, pixel_height],
        dtype=np.float32,
    )
    offset = np.array([rasterized_x_offset, rasterized_y_offset], dtype=np.int32)

    return new_vector_extent, rasterized_size, offset


@jit(nopython=True, nogil=True, fastmath=True, inline="always")
def get_intersection(extent1, extent2):
    one_bottomLeftX = extent1[0]
    one_topRightX = extent1[1]
    one_bottomLeftY = extent1[2]
    one_topRightY = extent1[3]

    two_bottomLeftX = extent2[0]
    two_topRightX = extent2[1]
    two_bottomLeftY = extent2[2]
    two_topRightY = extent2[3]

    if two_bottomLeftX > one_topRightX:  # Too far east
        return np.array([0, 0, 0, 0], dtype=np.float32)
    elif two_bottomLeftY > one_topRightY:  # Too far north
        return np.array([0, 0, 0, 0], dtype=np.float32)
    elif two_topRightX < one_bottomLeftX:  # Too far west
        return np.array([0, 0, 0, 0], dtype=np.float32)
    elif two_topRightY < one_bottomLeftY:  # Too far south
        return np.array([0, 0, 0, 0], dtype=np.float32)
    else:

        x_min = (
            one_bottomLeftX if one_bottomLeftX > two_bottomLeftX else two_bottomLeftX
        )
        x_max = one_topRightX if one_topRightX < two_topRightX else two_topRightX
        y_min = (
            one_bottomLeftY if one_bottomLeftY > two_bottomLeftY else two_bottomLeftY
        )
        y_max = one_topRightY if one_topRightY < two_topRightY else two_topRightY

        return np.array([x_min, x_max, y_min, y_max], dtype=np.float32)


@jit(nopython=True, nogil=True, fastmath=True, inline="always")
def get_extent(raster_transform, raster_size):
    bottomRightX = raster_transform[0] + (raster_size[0] * raster_transform[1])
    bottomRightY = raster_transform[3] + (raster_size[1] * raster_transform[5])

    return np.array(
        [raster_transform[0], bottomRightX, bottomRightY, raster_transform[3]],
        dtype=np.float32,
    )


def zonal_statistics(
    in_vector,
    output_vector=None,
    in_rasters=[],
    prefixes=[],
    stats=None,
):
    """
    ZonalStatistics.
    """
    default_stats = ["mean", "med", "std"]

    if stats is None:
        stats = default_stats

    if len(prefixes) != 0:
        if len(in_rasters) != len(prefixes):
            raise ValueError("Unable to parse prefixes.")

    if isinstance(in_rasters, list):
        if len(in_rasters) == 0:
            raise ValueError("List of rasters (in_rasters) is empty.")

    if len(stats) == 0:
        raise ValueError("Unable to parse statistics (stats).")

    # Read the raster meta:
    raster_metadata = raster_to_metadata(in_rasters[0])

    vector = None
    if output_vector is None:
        vector = open_vector(in_vector, writeable=True)
    else:
        vector = _vector_to_memory(in_vector)

    vector_metadata = _vector_to_metadata(vector)
    vector_layer = vector.GetLayer()

    # Check that projections match
    if not vector_metadata["projection_osr"].IsSame(raster_metadata["projection_osr"]):
        if output_vector is None:
            vector = reproject_vector(in_vector, in_rasters[0])
        else:
            vector_path = reproject_vector(
                in_vector, in_rasters[0], output_vector
            )
            vector = open_vector(vector_path, writeable=True)

        vector_metadata = _vector_to_metadata(vector)
        vector_layer = vector.GetLayer()

    vector_projection = vector_metadata["projection_osr"]
    raster_projection = raster_metadata["projection_osr"]

    # Read raster data in overlap
    raster_transform = np.array(raster_metadata["transform"], dtype=np.float32)
    raster_size = np.array(raster_metadata["size"], dtype=np.int32)

    raster_extent = get_extent(raster_transform, raster_size)

    vector_extent = np.array(vector_layer.GetExtent(), dtype=np.float32)
    overlap_extent = get_intersection(raster_extent, vector_extent)

    if overlap_extent is False:
        print("raster_extent: ", raster_extent)
        print("vector_extent: ", vector_extent)
        raise Exception("Vector and raster do not overlap!")

    (
        overlap_aligned_extent,
        overlap_aligned_rasterized_size,
        overlap_aligned_offset,
    ) = align_extent(raster_transform, overlap_extent, raster_size)
    overlap_transform = np.array(
        [
            overlap_aligned_extent[0],
            raster_transform[1],
            0,
            overlap_aligned_extent[3],
            0,
            raster_transform[5],
        ],
        dtype=np.float32,
    )
    overlap_size = overlap_size_calc(overlap_aligned_extent, raster_transform)

    # Loop the features
    vector_driver = ogr.GetDriverByName("Memory")
    vector_feature_count = vector_layer.GetFeatureCount()
    vector_layer.StartTransaction()

    # Create fields
    vector_layer_defn = vector_layer.GetLayerDefn()
    vector_field_counts = vector_layer_defn.GetFieldCount()
    vector_current_fields = []

    # Get current fields
    for i in range(vector_field_counts):
        vector_current_fields.append(vector_layer_defn.GetFieldDefn(i).GetName())

    # Add fields where missing
    for stat in stats:
        for i in range(len(in_rasters)):
            field_name = f"{prefixes[i]}{stat}"
            if field_name not in vector_current_fields:
                field_defn = ogr.FieldDefn(field_name, ogr.OFTReal)
                vector_layer.CreateField(field_defn)

    rasterized_features = []
    sizes = np.zeros((vector_feature_count, 4), dtype="float32")
    offsets = np.zeros((vector_feature_count, 2), dtype=np.int32)
    raster_data = None
    for raster_index, raster_value in enumerate(in_rasters):

        columns = {}
        for stat in stats:
            columns[prefixes[raster_index] + stat] = []

        fits_in_memory = True
        try:
            raster_data = raster_to_array(
                raster_value,
                pixel_offsets=[
                    overlap_aligned_offset[0],
                    overlap_aligned_offset[1],
                    overlap_aligned_rasterized_size[0],
                    overlap_aligned_rasterized_size[1],
                ],
            )
        except:
            fits_in_memory = False
            print("Raster does not fit in memory.. Doing IO for each feature.")

        for n in range(vector_feature_count):
            vector_feature = vector_layer.GetNextFeature()
            rasterized_vector = None

            if raster_index == 0:

                try:
                    vector_geom = vector_feature.GetGeometryRef()
                except:
                    vector_geom.Buffer(0)
                    Warning("Invalid geometry at : ", n)

                if vector_geom is None:
                    raise Exception("Invalid geometry. Could not fix.")

                feature_extent = vector_geom.GetEnvelope()

                # Create temp layer
                temp_vector_datasource = vector_driver.CreateDataSource(f"vector_{n}")
                temp_vector_layer = temp_vector_datasource.CreateLayer(
                    "temp_polygon", vector_projection, ogr.wkbPolygon
                )
                temp_vector_layer.CreateFeature(vector_feature.Clone())

                (
                    feature_aligned_extent,
                    feature_aligned_rasterized_size,
                    feature_aligned_offset,
                ) = align_extent(overlap_transform, feature_extent, overlap_size)

                rasterized_vector = None
                # rasterized_vector = rasterize_vector(
                #     temp_vector_layer,
                #     feature_aligned_extent,
                #     feature_aligned_rasterized_size,
                #     raster_projection,
                # )
                rasterized_features.append(rasterized_vector)

                offsets[n] = feature_aligned_offset
                sizes[n] = feature_aligned_rasterized_size

            if fits_in_memory is True:
                cropped_raster = raster_data[
                    offsets[n][1] : offsets[n][1] + int(sizes[n][1]),  # X
                    offsets[n][0] : offsets[n][0] + int(sizes[n][0]),  # Y
                ]
            else:
                cropped_raster = raster_to_array(
                    raster_value,
                    pixel_offsets=[
                        overlap_aligned_offset[0] + offsets[n][0],
                        overlap_aligned_offset[1] + offsets[n][1],
                        int(sizes[n][0]),
                        int(sizes[n][1]),
                    ],
                )

            if rasterized_features[n] is None:
                for stat in stats:
                    field_name = f"{prefixes[raster_index]}{stat}"
                    vector_feature.SetField(field_name, None)
            elif cropped_raster is None:
                for stat in stats:
                    field_name = f"{prefixes[raster_index]}{stat}"
                    vector_feature.SetField(field_name, None)
            else:
                raster_data_masked = np.ma.masked_array(
                    cropped_raster, mask=rasterized_features[n], dtype="float32"
                ).compressed()
                zonal_stats = calculate_array_stats(
                    raster_data_masked, stats
                )

                for index, stat in enumerate(stats):
                    field_name = f"{prefixes[raster_index]}{stat}"
                    vector_feature.SetField(field_name, float(zonal_stats[index]))

                vector_layer.SetFeature(vector_feature)

            progress(n, vector_feature_count, name=prefixes[raster_index])

        vector_layer.ResetReading()

    vector_layer.CommitTransaction()

    if output_vector is None:
        return vector

    return output_vector



"""
### Perform morphological operations on arrays and rasters.  ###
"""

# Standard library
import sys; sys.path.append("../../")

# Internal
from buteo.raster.convolution import get_kernel, convolve_array


def morph_erode(arr, size=5, spherical=False, distance_weight=False, distance_decay=0.2, distance_sigma=1.0):
    """ Erode an array by taking the local minimum. """

    _kernel, weights, offsets = get_kernel(size, arr.shape[-1], spherical=spherical, normalise=False, distance_weight=distance_weight, distance_decay=distance_decay, distance_sigma=distance_sigma)
    return convolve_array(arr, offsets, weights, method="min")


def morph_dilate(arr, size=5, spherical=False, distance_weight=False, distance_decay=0.2, distance_sigma=1.0):
    """ Dilate an array by taking the local maximum. """

    _kernel, weights, offsets = get_kernel(size, arr.shape[-1], spherical=spherical, normalise=False, distance_weight=distance_weight, distance_decay=distance_decay, distance_sigma=distance_sigma)
    return convolve_array(arr, offsets, weights, method="max")


def morph_open(arr, size=5, spherical=False, distance_weight=False, distance_decay=0.2, distance_sigma=1.0):
    """ Perform the open mortholigical operation on an array. """

    _kernel, weights, offsets = get_kernel(size, arr.shape[-1], spherical=spherical, normalise=False, distance_weight=distance_weight, distance_decay=distance_decay, distance_sigma=distance_sigma)
    eroded = convolve_array(arr, offsets, weights, method="min")
    return convolve_array(eroded, offsets, weights, method="max")


def morph_close(arr, size=5, spherical=False, distance_weight=False, distance_decay=0.2, distance_sigma=1.0):
    """ Perform the close morphological operation on an array. """

    _kernel, weights, offsets = get_kernel(size, arr.shape[-1], spherical=spherical, normalise=False, distance_weight=distance_weight, distance_decay=distance_decay, distance_sigma=distance_sigma)
    dilate = convolve_array(arr, offsets, weights, method="max")
    return convolve_array(dilate, offsets, weights, method="min")


def morph_tophat(arr, size=5, spherical=False, distance_weight=False, distance_decay=0.2, distance_sigma=1.0):
    """ Perform the top_hat morphological operation on the array. """

    _kernel, weights, offsets = get_kernel(size, arr.shape[-1], spherical=spherical, normalise=False, distance_weight=distance_weight, distance_decay=distance_decay, distance_sigma=distance_sigma)
    eroded = convolve_array(arr, offsets, weights, method="min")
    opened = convolve_array(eroded, offsets, weights, method="max")
    return arr - opened


def morph_bothat(arr, size=5, spherical=False, distance_weight=False, distance_decay=0.2, distance_sigma=1.0):
    """ Perform the bottom_hat morphological operation on the array. """

    _kernel, weights, offsets = get_kernel(size, arr.shape[-1], spherical=spherical, normalise=False, distance_weight=distance_weight, distance_decay=distance_decay, distance_sigma=distance_sigma)
    dilated = convolve_array(arr, offsets, weights, method="max")
    closed = convolve_array(dilated, offsets, weights, method="min")
    return closed - arr


def morph_difference(arr, size=5, spherical=False, distance_weight=False, distance_decay=0.2, distance_sigma=1.0):
    """ Perform the difference morphological operation on the array. """

    _kernel, weights, offsets = get_kernel(size, arr.shape[-1], spherical=spherical, normalise=False, distance_weight=distance_weight, distance_decay=distance_decay, distance_sigma=distance_sigma)
    erode = convolve_array(arr, offsets, weights, method="min")
    dilate = convolve_array(arr, offsets, weights, method="max")

    return dilate - erode
